package tetris;

import tetris.core.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;


@SuppressWarnings("serial")
public class Tetris extends JPanel implements KeyListener {

    protected Game game;
    private JLabel score;
    private BoardGUI boardView;

    public Tetris() {
        super();
        setPreferredSize(new Dimension(300, 600));
        
        game = new Game();
        score = new JLabel("score:");
        boardView = new BoardGUI();

        try {
            SwingUtilities.invokeAndWait(new Runnable() {
                public void run() {
                    layoutComponents();
                }
            });
        } catch (Exception e) {
            System.err.println("Error loading GUI.");
            System.exit(-1);
        }
        
        game.addScoreListener(new ScoreListener() {

            public void scoreChanged(ScoreEvent evt) {
                score.setText("score: " + evt.getScore());
                int delay = game.getDelay();
                int lines = evt.getLinesCleared();
                
                if (lines >= 50)
                    delay = 120;
                else if (lines >= 40)
                    delay = 150;
                else if (lines >= 30)
                    delay = 200;
                else if (lines >= 20)
                    delay = 300;
                else if (lines >= 10)
                    delay = 400;
                
                game.setDelay(delay);
            }
        });
        
        game.addBoardListener(boardView);
        addKeyListener(this);
    }

    private void layoutComponents() {
        JButton start = new JButton("start");
        start.addActionListener(new ActionListener() {

            public void actionPerformed(ActionEvent evt) {
                game.startGame();
                requestFocus();
            }
        });

        setLayout(new BorderLayout());
        add(boardView, BorderLayout.CENTER);
        add(start, BorderLayout.SOUTH);
        add(score, BorderLayout.NORTH);
    }

    public void keyPressed(KeyEvent e) {
        switch (e.getKeyCode()) {
        case KeyEvent.VK_LEFT:
            game.move(Piece.Move.LEFT);
            break;
        case KeyEvent.VK_RIGHT:
            game.move(Piece.Move.RIGHT);
            break;
        case KeyEvent.VK_UP:
            game.move(Piece.Move.ROTATE);
            break;
        case KeyEvent.VK_DOWN:
            game.move(Piece.Move.DOWN);
            break;
        case KeyEvent.VK_SPACE:
            game.move(Piece.Move.FALL);
            break;
        default:
            break;
        }
    }

    public void keyReleased(KeyEvent e) {
    }

    public void keyTyped(KeyEvent e) {
    }

    private class BoardGUI extends JPanel implements BoardListener {

        private Board boardRef;

        public void boardChanged(BoardEvent evt) {
            boardRef = evt.getSource();
            repaint();
        }

        public void paint(Graphics graphics) {
            final int width = getBounds().width;
            final int height = getBounds().height;

            // Set up double buffering.
            final Image image = createImage(width, height);
            final Graphics g = image.getGraphics();

            // Draw the board pieces if board not null.
            if (boardRef != null) {
                final int numCols = boardRef.getWidth();
                final int numRows = boardRef.getHeight();

                for (int cols = 0; cols < numCols; cols++) {
                    for (int rows = 0; rows < numRows; rows++) {
                        Piece.Type piecet = boardRef.getPieceTypeAt(cols, rows);

                        if (piecet != null) {
                            g.setColor(getPieceColor(piecet));
                            drawBlock(g, (cols * width / numCols) + 1, (rows
                                    * height / numRows) + 1,
                                    (width / numCols) - 1,
                                    (height / numRows) - 1);
                        }
                    }
                }
            }

            g.setColor(Color.RED);
            g.drawRect(0, 0, width - 1, height - 1);
            g.dispose();
            
            graphics.drawImage(image, 0, 0, width, height, null);
        }

        private void drawBlock(Graphics g, int x, int y, int width, int height) {
            g.fillRect(x, y, width, height);
            g.setColor(Color.BLACK);
            g.drawRect(x, y, width, height);
        }

        private Color getPieceColor(Piece.Type type) {
            Color color = null;
            
            switch (type) {
            case L:
                color = new Color(24, 105, 198);
                break;
            case J:
                color = new Color(206, 56, 173);
                break;
            case I:
                color = new Color(41, 40, 204);
                break;
            case Z:
                color = new Color(212, 0, 0);
                break;
            case S:
                color = new Color(82, 154, 16);
                break;
            case O:
                color = new Color(123, 121, 123);
                break;
            case T:
                color = new Color(156, 142, 8);
                break;
            }
            
            return color;
        }
    }
}
