package tetris.core;

import java.util.*;
import java.awt.Point;


public class Board {

    private int width, height;
    private Piece.Type[][] cells;

    private ArrayList<BoardListener> listeners;

    public Board(int width, int height) {
        this.width = width;
        this.height = height;
        listeners = new ArrayList<BoardListener>();
        resetBoard();
    }

    private void resetBoard() {
        cells = new Piece.Type[height][width];
        for (int i = 0; i < height; i++)
            for (int j = 0; j < width; j++)
                cells[i][j] = null;
    }

    public int getWidth() {
        return width;
    }

    public int getHeight() {
        return height;
    }

    public void setWidth(int w) {
        width = w;
        resetBoard();
    }

    public void setHeight(int h) {
        height = h;
        resetBoard();
    }

    public void setDimension(int w, int h) {
        width = w;
        height = h;
        resetBoard();
    }

    // where (0, 0) is the top left corner of the board
    public Piece.Type getPieceTypeAt(int x, int y) {
        return cells[y][x];
    }

    public boolean addPiece(Piece piece, boolean notify) {
        if (piece == null || !willFit(piece))
            return false;

        Point pcenter = piece.getCenter();
        for (Point p : piece.getBody())
            cells[pcenter.y + p.y][pcenter.x + p.x] = piece.getType();

        if (notify) {
            fireBoardEvent();
        }
        return true;
    }

    protected boolean willFit(Piece piece) {
        Point pcenter = piece.getCenter();
        for (Point p : piece.getBody())
            try {
                if (cells[pcenter.y + p.y][pcenter.x + p.x] != null)
                    return false;
            } catch (ArrayIndexOutOfBoundsException e) {
                return false;
            }
        return true;
    }

    public void removePiece(Piece piece) {
        if (piece == null)
            return;

        Point pcenter = piece.getCenter();
        for (Point p : piece.getBody())
            cells[pcenter.y + p.y][pcenter.x + p.x] = null;
    }

    public void removeRow(int row) {
        // shift all rows above row down one row
        for (int r = row; r > 0; r--)
            for (int c = 0; c < width; c++)
                cells[r][c] = cells[r - 1][c];
        // set topmost row to all blanks
        for (int c = 0; c < width; c++)
            cells[0][c] = null;

        fireBoardEvent();
    }

    public void addBoardListener(BoardListener listener) {
        listeners.add(listener);
    }

    public void removeBoardListener(BoardListener listener) {
        listeners.remove(listener);
    }

    private void fireBoardEvent() {
        BoardEvent event = new BoardEvent(this);
        for (BoardListener listener : listeners)
            listener.boardChanged(event);
    }

    // for unit testing or debugging purposes
    public String toString() {
        StringBuffer sbuf = new StringBuffer();
        for (int r = 0; r < height; r++) {
            for (int c = 0; c < width; c++) {
                if (cells[r][c] == null)
                    sbuf.append('.');
                else
                    sbuf.append(cells[r][c]);
            }
            if (r < height - 1)
                sbuf.append('\n');
        }
        return sbuf.toString();
    }

    // unit test:
    public static void main(String[] args) {
        boolean success = false;
        Piece piece = null;
        Board board = new Board(10, 10);
        System.out.println("creating a 10 by 10 Board ... done");
        piece = Piece.getRandomPiece(board);
        piece.setCenter(new Point(4, 0));
        success = board.addPiece(piece, false);
        System.out.println("adding a random piece at (4, 0) ... "
                + (success ? "success" : "failed"));
        success = piece.move(Piece.Move.RIGHT);
        System.out.println("moving the piece RIGHT once ... "
                + (success ? "success" : "failed"));
        success = piece.move(Piece.Move.DOWN);
        success = success && piece.move(Piece.Move.DOWN);
        System.out.println("moving the piece DOWN twice ... "
                + (success ? "success" : "failed"));
        success = piece.move(Piece.Move.LEFT);
        System.out.println("moving the piece LEFT once ... "
                + (success ? "success" : "failed"));
        success = piece.move(Piece.Move.ROTATE);
        System.out.println("moving the piece to ROTATE once ... "
                + (success ? "success" : "failed"));
        success = piece.move(Piece.Move.FALL);
        System.out.println("moving the piece to FALL ... "
                + (success ? "success" : "failed"));
        success = true;
        for (int i = 0; i < 10; i++)
            success = success && piece.move(Piece.Move.RIGHT);
        System.out.println("moving the piece RIGHT ten times ... "
                + (success ? "success" : "failed"));
        board.removeRow(9);
        System.out.println("removing last row ... ");

        System.out.println();
        System.out.println(board);
    }
}
