package tetris.core;

import java.util.ArrayList;
import java.awt.Point;


public class Game {

    private Board board;
    private Piece curPiece, nxtPiece;
    private boolean isPlaying, isPaused;
    private int delay, score, linesCleared;
    private Thread gameThread;

    private ArrayList<GameListener> gameListeners;
    private ArrayList<ScoreListener> scoreListeners;

    public Game() {
        this(20, 10);
    }

    public Game(int row, int col) {
        board = new Board(col, row);
        isPlaying = false;

        gameListeners = new ArrayList<GameListener>();
        scoreListeners = new ArrayList<ScoreListener>();
    }

    public void startGame() {
        if (!isPlaying) {
            curPiece = null;
            nxtPiece = Piece.getRandomPiece(board);
            isPlaying = true;
            isPaused = false;
            score = linesCleared = 0;
            delay = 500;

            fireScoreEvent();
            fireGameEvent(new GameEvent(this, GameEvent.State.START));

            gameThread = new GameThread();
            gameThread.start();
        }
    }

    public void stopGame() {
        isPlaying = false;
        fireGameEvent(new GameEvent(this, GameEvent.State.END));
        gameThread = null;
    }

    public int getNumRows() {
        return board.getHeight();
    }

    public int getNumCols() {
        return board.getWidth();
    }

    public Piece getCurrentPiece() {
        return curPiece;
    }

    public Piece getNextPiece() {
        return nxtPiece;
    }

    public boolean isPlaying() {
        return isPlaying;
    }

    public boolean isPaused() {
        return isPaused;
    }

    public int getScore() {
        return score;
    }

    public int getLinesCleared() {
        return linesCleared;
    }

    public int getDelay() {
        return delay;
    }

    public void setDelay(int delay) {
        this.delay = delay;
    }

    public void setPaused(boolean paused) {
        if (isPlaying)
            isPaused = paused;
    }

    public boolean move(Piece.Move direction) {
        boolean result = false;

        if (isPlaying && !isPaused && curPiece != null) {
            result = curPiece.move(direction);
            if (direction == Piece.Move.DOWN || direction == Piece.Move.FALL) {
                if (!result) // FALL always returns false, and DOWN when a piece
                    curPiece = null; // is moved to the most bottom it can go
            }
        }

        return result;
    }

    public void addBoardListener(BoardListener bl) {
        board.addBoardListener(bl);
    }

    public void addGameListener(GameListener gl) {
        gameListeners.add(gl);
    }

    public void addScoreListener(ScoreListener sl) {
        scoreListeners.add(sl);
    }

    public void removeBoardListener(BoardListener bl) {
        board.removeBoardListener(bl);
    }

    public void removeGameListener(GameListener gl) {
        gameListeners.remove(gl);
    }

    public void removeScoreListener(ScoreListener sl) {
        scoreListeners.remove(sl);
    }

    private void fireScoreEvent() {
        ScoreEvent evt = new ScoreEvent(score, linesCleared);
        for (ScoreListener sl : scoreListeners)
            sl.scoreChanged(evt);
    }

    private void fireGameEvent(GameEvent evt) {
        for (GameListener sl : gameListeners)
            sl.gameStateChanged(evt);
    }

    private class GameThread extends Thread {

        public void run() {
            while (isPlaying) {
                if (!isPaused) {
                    if (curPiece == null) {
                        int filledRows = 0;
                        for (int r = board.getHeight() - 1; r >= 0; r--) {
                            boolean filled = true;
                            // check if this row r is all filled
                            for (int c = 0; c < board.getWidth(); c++) {
                                if (board.getPieceTypeAt(c, r) == null) {
                                    filled = false;
                                    break;
                                }
                            }
                            // remove the filled row
                            if (filled) {
                                board.removeRow(r);
                                r++;
                                filledRows++;
                                linesCleared++;
                            }
                        }
                        // at this point, all filled rows have been removed
                        if (filledRows > 0) {
                            score += filledRows * filledRows * 10;
                            fireScoreEvent();
                        }
                        // randomize next piece
                        curPiece = nxtPiece;
                        nxtPiece = Piece.getRandomPiece(board);
                        // place it on the board
                        curPiece.setCenter(new Point(board.getWidth() / 2, 1));
                        if (!board.addPiece(curPiece, true)) {
                            System.err.println("tried to add piece: "
                                    + curPiece);
                            stopGame();
                            break;
                        }
                    } else {
                        // drop the current piece DOWN
                        move(Piece.Move.DOWN);
                    }
                }

                if (curPiece != null)
                    try {
                        sleep(delay);
                    } catch (InterruptedException ex) {
                        System.err.println("GameThread interrupted: " + ex);
                    }
            }
        }
    }
}
