package tetris.core;

import java.awt.Point;
import java.util.Random;


public class Piece {

    public enum Type {
        L(4, new Point[] { new Point(0, 0), new Point(0, 1), new Point(1, 0), new Point(2, 0) }), 
        J(4, new Point[] { new Point(0, 0), new Point(-1, 0), new Point(-2, 0), new Point(0, 1) }), 
        S(2, new Point[] { new Point(0, 0), new Point(1, 0), new Point(0, 1), new Point(-1, 1) }), 
        O(0, new Point[] { new Point(0, 0), new Point(1, 0), new Point(0, 1), new Point(1, 1) }), 
        I(2, new Point[] { new Point(0, 0), new Point(-1, 0), new Point(1, 0), new Point(2, 0) }), 
        Z(2, new Point[] { new Point(0, 0), new Point(-1, 0), new Point(0, 1), new Point(1, 1) }), 
        T(4, new Point[] { new Point(0, 0), new Point(-1, 0), new Point(1, 0), new Point(0, 1) });

        private final int rotationFactor;
        private final Point[] body;

        private Type(int rotf, Point[] bod) {
            rotationFactor = rotf;
            body = bod;
        }

        public int getRotationFactor() {
            return rotationFactor;
        }

        public Point[] getBody() {
            return body;
        }
    }

    public enum Move {
        LEFT, RIGHT, DOWN, FALL, ROTATE;

        public boolean move(Piece piece) {
            Board boardRef = piece.getBoardRef();
            boolean result = true;

            switch (this) { // TODO
            case LEFT:
                boardRef.removePiece(piece);
                piece.moveLeft();

                if (!boardRef.addPiece(piece, true)) {
                    piece.moveRight();
                    boardRef.addPiece(piece, true);
                    result = false;
                }

                return result;
            case RIGHT:
                boardRef.removePiece(piece);
                piece.moveRight();

                if (!boardRef.addPiece(piece, true)) {
                    piece.moveLeft();
                    boardRef.addPiece(piece, true);
                    result = false;
                }

                return result;
            case DOWN:
                boardRef.removePiece(piece);
                piece.moveDown();

                if (!boardRef.addPiece(piece, true)) {
                    piece.moveUp();
                    boardRef.addPiece(piece, true);
                    result = false;
                }

                return result;
            case ROTATE:
                boardRef.removePiece(piece);
                piece.rotateClockwise();

                if (!boardRef.addPiece(piece, true)) {
                    piece.rotateClockwise();
                    piece.rotateClockwise();
                    piece.rotateClockwise();
                    boardRef.addPiece(piece, true);
                    result = false;
                }

                return result;
            case FALL:
                boolean loop = true;

                while (loop) {
                    boardRef.removePiece(piece);
                    piece.moveDown();

                    if (!boardRef.addPiece(piece, false)) {
                        piece.moveUp();
                        boardRef.addPiece(piece, true);
                        loop = false;
                        result = false;
                    }
                }

                return result;
            }
            throw new AssertionError("Unknown move: " + this);
        }
    }

    private Board board; // reference to the Board this Piece belongs to
    private Type type;
    private Point center; // on the board
    private Point[] body; // relative to piece's `center'
    private int rotation;

    public static final Random generator = new Random();

    public Piece(Type type, Board boardRef) {
        this.type = type;
        this.board = boardRef;
        initPiece();
    }

    public Piece(Type type, Board boardRef, Point iniCenter) {
        this(type, boardRef);
        center = iniCenter;
    }

    private void initPiece() {
        this.body = type.getBody().clone();
        this.center = new Point();
        this.rotation = 0;
    }

    public Point getCenter() {
        return center;
    }

    public Point[] getBody() {
        return body;
    }

    public Type getType() {
        return type;
    }

    public Board getBoardRef() {
        return board;
    }

    public void setCenter(Point p) {
        this.center = p;
    }

    public void setBody(Point[] d) {
        if (d != null)
            this.body = d;
    }

    public boolean move(Move direction) {
        return direction.move(this);
    }

    private void moveUp() {
        center.y--;
    }

    private void moveDown() {
        center.y++;
    }

    private void moveLeft() {
        center.x--;
    }

    private void moveRight() {
        center.x++;
    }

    private void rotateClockwise() {
        int rotf = type.getRotationFactor();
        if (rotf > 0) { // ignore angularly symmetric shapes
            rotation++;
            // for shapes with rotation factor 2,
            // rotate clockwise and anticlockwise alternatingly
            if (rotf == 2 && rotation == 2) {
                // rotate anti-clockwise
                doRotateCW();
                doRotateCW();
                doRotateCW();
            } else
                doRotateCW();
            rotation = rotation % rotf;
        }
    }

    private void doRotateCW() {
        for (int i = 0; i < body.length; i++) {
            Point oldpt = body[i];
            Point newpt = new Point(-oldpt.y, oldpt.x);
            body[i] = newpt;
        }
    }

    public static Type getRandomType() {
        int type = generator.nextInt(7);

        switch (type) {
        case 0:
            return Type.L;
        case 1:
            return Type.J;
        case 2:
            return Type.S;
        case 3:
            return Type.O;
        case 4:
            return Type.I;
        case 5:
            return Type.Z;
        case 6:
            return Type.T;
        default: // shouldn't be reached
            throw new AssertionError("Unknown type: " + type);
        }
    }

    public static Piece getRandomPiece(Board boardRef) {
        return new Piece(getRandomType(), boardRef);
    }

    public String toString() {
        char[][] fields = new char[5][5];
        for (int i = 0; i < 5; i++)
            for (int j = 0; j < 5; j++)
                fields[i][j] = '.';
        for (Point p : getBody())
            fields[p.y + 2][p.x + 2] = 'o';
        fields[2][2] = 'O';

        StringBuffer repr = new StringBuffer();
        repr.append("Piece.Type=" + getType() + "; center=" + getCenter()
                + "; rotation=" + rotation + ";\n");
        for (int i = 0; i < 5; i++) {
            for (int j = 0; j < 5; j++)
                repr.append(fields[i][j]);
            if (i < 4)
                repr.append('\n');
        }
        return repr.toString();
    }

    // unit test:
    public static void main(String[] args) {
        Piece[] pieces = new Piece[7];
        pieces[0] = new Piece(Type.S, null);
        pieces[1] = new Piece(Type.Z, null);
        pieces[2] = new Piece(Type.T, null);
        pieces[3] = new Piece(Type.L, null);
        pieces[4] = new Piece(Type.J, null);
        pieces[5] = new Piece(Type.I, null);
        pieces[6] = new Piece(Type.O, null);

        /*
         * for (Piece p : pieces) { System.out.println("Piece:");
         * System.out.println(p); }
         */

        Piece s = pieces[0];
        Piece t = pieces[2];
        Piece i = pieces[5];
        Piece o = pieces[6];

        System.out.println("S piece:\n" + s);
        s.rotateClockwise();
        System.out.println("S rotated once:\n" + s);
        s.rotateClockwise();
        System.out.println("S rotated twice:\n" + s);
        s.rotateClockwise();
        System.out.println("S rotated thrice:\n" + s);

        System.out.println("T piece:\n" + t);
        t.rotateClockwise();
        System.out.println("T rotated once:\n" + t);
        t.rotateClockwise();
        System.out.println("T rotated twice:\n" + t);
        t.rotateClockwise();
        System.out.println("T rotated thrice:\n" + t);
        t.rotateClockwise();
        System.out.println("T rotated four times:\n" + t);

        System.out.println("I piece:\n" + i);
        i.rotateClockwise();
        System.out.println("I rotated once:\n" + i);
        i.rotateClockwise();
        System.out.println("I rotated twice:\n" + i);
        i.rotateClockwise();
        System.out.println("I rotated thrice:\n" + i);

        System.out.println("O piece:\n" + o);
        o.rotateClockwise();
        System.out.println("O rotated once:\n" + o);
        o.rotateClockwise();
        System.out.println("I rotated twice:\n" + o);
    }
}
